package sun.rmi.server;

import java.util.Collections;
import java.util.WeakHashMap;
import sun.security.action.GetBooleanAction;
import java.rmi.server.LogStream;
import sun.security.action.GetPropertyAction;
import java.security.NoSuchAlgorithmException;
import java.io.IOException;
import java.io.DataOutputStream;
import java.security.DigestOutputStream;
import java.security.MessageDigest;
import java.io.ByteArrayOutputStream;
import java.rmi.server.SkeletonNotFoundException;
import java.rmi.server.Skeleton;
import java.lang.reflect.InvocationTargetException;
import java.rmi.server.RemoteStub;
import java.rmi.RemoteException;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.security.AccessController;
import java.lang.reflect.Proxy;
import java.lang.reflect.InvocationHandler;
import java.security.PrivilegedAction;
import java.rmi.server.RemoteObjectInvocationHandler;
import java.rmi.StubNotFoundException;
import java.rmi.Remote;
import java.rmi.server.RemoteRef;
import java.util.Map;
import sun.rmi.runtime.Log;

public final class Util
{
    static final int logLevel;
    public static final Log serverRefLog;
    private static final boolean ignoreStubClasses;
    private static final Map<Class<?>, Void> withoutStubs;
    private static final Class<?>[] stubConsParamTypes;
    
    private Util() {
    }
    
    public static Remote createProxy(Class<?> clazz, RemoteRef ref, boolean b) throws StubNotFoundException {
        Class<Object> remoteClass;
        try {
            remoteClass = getRemoteClass(clazz);
        }
        catch (ClassNotFoundException ex) {
            throw new StubNotFoundException("object does not implement a remote interface: " + clazz.getName());
        }
        if (b || (!Util.ignoreStubClasses && stubClassExists(remoteClass))) {
            return createStub(remoteClass, ref);
        }
        final ClassLoader classLoader = clazz.getClassLoader();
        final Class[] remoteInterfaces = getRemoteInterfaces(clazz);
        final RemoteObjectInvocationHandler remoteObjectInvocationHandler = new RemoteObjectInvocationHandler(ref);
        try {
            return (Remote)AccessController.doPrivileged(new PrivilegedAction<Remote>() {
                @Override
                public Remote run() {
                    return (Remote)Proxy.newProxyInstance(classLoader, remoteInterfaces, remoteObjectInvocationHandler);
                }
            });
        }
        catch (IllegalArgumentException ex2) {
            throw new StubNotFoundException("unable to create proxy", ex2);
        }
    }
    
    private static boolean stubClassExists(Class<?> clazz) {
        if (!Util.withoutStubs.containsKey(clazz)) {
            try {
                Class.forName(clazz.getName() + "_Stub", false, clazz.getClassLoader());
                return true;
            }
            catch (ClassNotFoundException ex) {
                Util.withoutStubs.put(clazz, null);
            }
        }
        return false;
    }
    
    private static Class<?> getRemoteClass(Class<?> superclass) throws ClassNotFoundException {
        while (superclass != null) {
            Class[] interfaces = superclass.getInterfaces();
            int i = interfaces.length - 1;
            while (i >= 0) {
                if (Remote.class.isAssignableFrom(interfaces[i])) {
                    return superclass;
                }
                --i;
            }
            superclass = superclass.getSuperclass();
        }
        throw new ClassNotFoundException("class does not implement java.rmi.Remote");
    }
    
    private static Class<?>[] getRemoteInterfaces(Class<?> clazz) {
        ArrayList<Class<?>> list = new ArrayList();
        getRemoteInterfaces(list, clazz);
        return (Class[])list.toArray(new Class[list.size()]);
    }
    
    private static void getRemoteInterfaces(ArrayList<Class<?>> list, Class<?> clazz) {
        Class<Object> superclass = clazz.getSuperclass();
        if (superclass != null) {
            getRemoteInterfaces(list, superclass);
        }
        Class[] interfaces = clazz.getInterfaces();
        for (int i = 0; i < interfaces.length; ++i) {
            Class<Object> clazz2 = interfaces[i];
            if (Remote.class.isAssignableFrom(clazz2) && !list.contains(clazz2)) {
                Method[] methods = clazz2.getMethods();
                for (int j = 0; j < methods.length; ++j) {
                    checkMethod(methods[j]);
                }
                list.add(clazz2);
            }
        }
    }
    
    private static void checkMethod(Method obj) {
        Class[] exceptionTypes = obj.getExceptionTypes();
        int i = 0;
        while (i < exceptionTypes.length) {
            if (exceptionTypes[i].isAssignableFrom(RemoteException.class)) {
                return;
            }
            ++i;
        }
        throw new IllegalArgumentException("illegal remote method encountered: " + obj);
    }
    
    private static RemoteStub createStub(Class<?> clazz, RemoteRef remoteRef) throws StubNotFoundException {
        String string = clazz.getName() + "_Stub";
        try {
            return (RemoteStub)Class.forName(string, false, clazz.getClassLoader()).getConstructor(Util.stubConsParamTypes).newInstance(new Object[] { remoteRef });
        }
        catch (ClassNotFoundException ex) {
            throw new StubNotFoundException("Stub class not found: " + string, ex);
        }
        catch (NoSuchMethodException ex2) {
            throw new StubNotFoundException("Stub class missing constructor: " + string, ex2);
        }
        catch (InstantiationException ex3) {
            throw new StubNotFoundException("Can't create instance of stub class: " + string, ex3);
        }
        catch (IllegalAccessException ex4) {
            throw new StubNotFoundException("Stub class constructor not public: " + string, ex4);
        }
        catch (InvocationTargetException ex5) {
            throw new StubNotFoundException("Exception creating instance of stub class: " + string, ex5);
        }
        catch (ClassCastException ex6) {
            throw new StubNotFoundException("Stub class not instance of RemoteStub: " + string, ex6);
        }
    }
    
    static Skeleton createSkeleton(Remote remote) throws SkeletonNotFoundException {
        Class<Object> remoteClass;
        try {
            remoteClass = getRemoteClass(remote.getClass());
        }
        catch (ClassNotFoundException ex) {
            throw new SkeletonNotFoundException("object does not implement a remote interface: " + remote.getClass().getName());
        }
        String string = remoteClass.getName() + "_Skel";
        try {
            return (Skeleton)Class.forName(string, false, remoteClass.getClassLoader()).newInstance();
        }
        catch (ClassNotFoundException ex2) {
            throw new SkeletonNotFoundException("Skeleton class not found: " + string, ex2);
        }
        catch (InstantiationException ex3) {
            throw new SkeletonNotFoundException("Can't create skeleton: " + string, ex3);
        }
        catch (IllegalAccessException ex4) {
            throw new SkeletonNotFoundException("No public constructor: " + string, ex4);
        }
        catch (ClassCastException ex5) {
            throw new SkeletonNotFoundException("Skeleton not of correct class: " + string, ex5);
        }
    }
    
    public static long computeMethodHash(Method method) {
        long n = 0L;
        ByteArrayOutputStream stream = new ByteArrayOutputStream(127);
        try {
            MessageDigest instance = MessageDigest.getInstance("SHA");
            DataOutputStream dataOutputStream = new DataOutputStream(new DigestOutputStream(stream, instance));
            String methodNameAndDescriptor = getMethodNameAndDescriptor(method);
            if (Util.serverRefLog.isLoggable(Log.VERBOSE)) {
                Util.serverRefLog.log(Log.VERBOSE, "string used for method hash: \"" + methodNameAndDescriptor + "\"");
            }
            dataOutputStream.writeUTF(methodNameAndDescriptor);
            dataOutputStream.flush();
            byte[] digest = instance.digest();
            for (int i = 0; i < Math.min(8, digest.length); ++i) {
                n += (long)(digest[i] & 0xFF) << i * 8;
            }
        }
        catch (IOException ex) {
            n = -1L;
        }
        catch (NoSuchAlgorithmException ex2) {
            throw new SecurityException(ex2.getMessage());
        }
        return n;
    }
    
    private static String getMethodNameAndDescriptor(Method method) {
        StringBuffer sb = new StringBuffer(method.getName());
        sb.append('(');
        Class[] parameterTypes = method.getParameterTypes();
        for (int i = 0; i < parameterTypes.length; ++i) {
            sb.append(getTypeDescriptor(parameterTypes[i]));
        }
        sb.append(')');
        Class<Object> returnType = method.getReturnType();
        if (returnType == Void.TYPE) {
            sb.append('V');
        }
        else {
            sb.append(getTypeDescriptor(returnType));
        }
        return sb.toString();
    }
    
    private static String getTypeDescriptor(Class<?> obj) {
        if (obj.isPrimitive()) {
            if (obj == Integer.TYPE) {
                return "I";
            }
            if (obj == Boolean.TYPE) {
                return "Z";
            }
            if (obj == Byte.TYPE) {
                return "B";
            }
            if (obj == Character.TYPE) {
                return "C";
            }
            if (obj == Short.TYPE) {
                return "S";
            }
            if (obj == Long.TYPE) {
                return "J";
            }
            if (obj == Float.TYPE) {
                return "F";
            }
            if (obj == Double.TYPE) {
                return "D";
            }
            if (obj == Void.TYPE) {
                return "V";
            }
            throw new Error("unrecognized primitive type: " + obj);
        }
        else {
            if (obj.isArray()) {
                return obj.getName().replace('.', '/');
            }
            return "L" + obj.getName().replace('.', '/') + ";";
        }
    }
    
    public static String getUnqualifiedName(Class<?> clazz) {
        String name = clazz.getName();
        return name.substring(name.lastIndexOf(46) + 1);
    }
    
    static {
        logLevel = LogStream.parseLevel((String)AccessController.doPrivileged(new GetPropertyAction("sun.rmi.server.logLevel")));
        serverRefLog = Log.getLog("sun.rmi.server.ref", "transport", Util.logLevel);
        ignoreStubClasses = (boolean)AccessController.doPrivileged(new GetBooleanAction("java.rmi.server.ignoreStubClasses"));
        withoutStubs = Collections.synchronizedMap(new WeakHashMap(11));
        stubConsParamTypes = new Class[] { RemoteRef.class };
    }
}
